body
{
  color: #000000;
  background-color: #FFFFFF;
}

/* Link formats: use maroon w/underlines */
a:link
{
  color: navy;
  text-decoration: underline;
}
a:visited
{
  color: maroon;
  text-decoration: underline;
}
a:active
{
  color: navy;
  text-decoration: underline;
}

/*** TD formats ***/
td
{
  font-family: sans-serif;
}
td.title
{
  text-align: center;
  padding-bottom: 10px;
  font-size: 20pt;
  font-weight: bold;
}

/* TD Header Information */
td.headerName
{
  text-align: right;
  color: black;
  padding-right: 6px;
  font-weight: bold;
  vertical-align: top;
  white-space: nowrap;
}
td.headerValue
{
  text-align: left;
  color: blue;
  font-weight: bold;
  white-space: nowrap;
}
td.headerTableEntry
{
  text-align: right;
  color: black;
  font-weight: bold;
  white-space: nowrap;
  padding-left: 12px;
  padding-right: 4px;
  background-color: LightBlue;
}
td.headerValueLeg
{
  text-align: left;
  color: black;
  font-size: 80%;
  white-space: nowrap;
  padding-left: 10px;
  padding-right: 10px;
  padding-top: 2px;
}

/* Color of horizontal ruler */
td.hr
{
  background-color: navy;
  height:3px;
}
/* Footer format */
td.footer
{
  text-align: center;
  padding-top: 3px;
  font-family: sans-serif;
}

/* Coverage Table */

td.coverTableHead
{
  text-align: center;
  color: white;
  background-color: SteelBlue;
  font-family: sans-serif;
  font-size: 120%;
  white-space: nowrap;
  padding-left: 4px;
  padding-right: 4px;
}
td.coverFile
{
  text-align: left;
  padding-left: 10px;
  padding-right: 20px;
  color: black;
  background-color: LightBlue;
  font-family: monospace;
  font-weight: bold;
  font-size: 110%;
}
td.coverBar
{
  padding-left: 10px;
  padding-right: 10px;
  background-color: LightBlue;
}
td.coverBarOutline
{
  background-color: white;
}
td.coverValue
{
  padding-top: 2px;
  text-align: right;
  padding-left: 10px;
  padding-right: 10px;
  font-family: sans-serif;
  white-space: nowrap;
  font-weight: bold;
}

/* Link Details */
a.detail:link
{
  color: #B8D0FF;
  font-size:80%;
}
a.detail:visited
{
  color: #B8D0FF;
  font-size:80%;
}
a.detail:active
{
  color: #FFFFFF;
  font-size:80%;
}

.graphcont{
    color:#000;
    font-weight:700;
    float:left
}

.graph{
    float:left;
    background-color: white;
    position:relative;
    width:280px;
    padding:0
}

.graph .bar{
    display:block;
    position:relative;
    border:black 1px solid;
    text-align:center;
    color:#fff;
    height:10px;
    font-family:Arial,Helvetica,sans-serif;
    font-size:12px;
    line-height:1.9em
}

.graph .bar span{
    position:absolute;
    left:1em
}

td.coveredLine,
span.coveredLine
{
    background-color: {{covered_color}}!important;
}

td.uncoveredLine,
span.uncoveredLine
{
    background-color: {{uncovered_color}}!important;
}

.linebranch, .linecount
{
    border-right: 1px gray solid;
    background-color: lightgray;
}

span.takenBranch
{
    color: {{takenBranch_color}}!important;
    cursor: help;
}

span.notTakenBranch
{
    color: {{notTakenBranch_color}}!important;
    cursor: help;
}

.src
{
    padding-left: 12px;
}

.srcHeader,
span.takenBranch,
span.notTakenBranch
{
    font-family: monospace;
    font-weight: bold;
}

pre
{
    height : 15px;
    margin-top: 0;
    margin-bottom: 0;
}

.lineno
{
    background-color: #EFE383;
    border-right: 1px solid #BBB15F;
}
