
<html>

<head>
  <meta http-equiv="Content-Type" content="text/html; charset={{ENC}}"/>
  <title>{{HEAD}}</title>
  <style media="screen" type="text/css">

    {{CSS | indent}}

  </style>
</head>

<body>

  <table width="100%" border="0" cellspacing="0" cellpadding="0">
    <tr><td class="title">GCC Code Coverage Report</td></tr>
    <tr><td class="hr"></td></tr>

    <tr>
      <td width="100%">
        <table cellpadding="1" border="0" width="100%">
          <tr>
            <td width="10%" class="headerName">Directory:</td>
            <td width="35%" class="headerValue">{{DIRECTORY}}</td>
            <td width="5%"></td>
            <td width="15%"></td>
            <td width="10%" class="headerValue" style="text-align:right;">Exec</td>
            <td width="10%" class="headerValue" style="text-align:right;">Total</td>
            <td width="15%" class="headerValue" style="text-align:right;">Coverage</td>
          </tr>
          <tr>
            <td class="headerName">File:</td>
            <td class="headerValue">{{FILENAME}}</td>
            <td></td>
            <td class="headerName">Lines:</td>
            <td class="headerTableEntry">{{LINES_EXEC}}</td>
            <td class="headerTableEntry">{{LINES_TOTAL}}</td>
            <td class="headerTableEntry" style="background-color:{{LINES_COLOR}}">{{LINES_COVERAGE}} %</td>
          </tr>
          <tr>
            <td class="headerName">Date:</td>
            <td class="headerValue">{{DATE}}</td>
            <td></td>
            <td class="headerName">Branches:</td>
            <td class="headerTableEntry">{{BRANCHES_EXEC}}</td>
            <td class="headerTableEntry">{{BRANCHES_TOTAL}}</td>
            <td class="headerTableEntry" style="background-color:{{BRANCHES_COLOR}}">{{BRANCHES_COVERAGE}} %</td>
          </tr>
        </table>
      </td>
    </tr>

    <tr><td class="hr"></td></tr>
  </table>

  <br>
  <table cellspacing="0" cellpadding="1">
    <tr>
      <td width="5%" align="right" class="srcHeader">Line</td>
      <td width="5%" align="right" class="srcHeader">Branch</td>
      <td width="5%" align="right" class="srcHeader">Exec</td>
      <td width="75%" align="left" class="srcHeader src">Source</td>
    </tr>

    {% for row in ROWS %}

    <tr>
    <td align="right" class="lineno"><pre>{{row.lineno}}</pre></td>
    <td align="right" class="linebranch">
      {%- for branch in row.linebranch -%}
      <span class="{{branch.class}}" title="{{branch.message}}">{{branch.symbol}}</span>
      {%- if branch.wrap %}<br/>{% endif -%}
      {%- endfor -%}
    </td>
    <td align="right" class="linecount {{row.covclass}}"><pre>{{row.linecount}}</pre></td>
    <td align="left" class="src {{row.covclass}}"><pre>{{row.source}}</pre></td>
    </tr>
    {% endfor %}

  </table>
  <br>

  <table width="100%" border="0" cellspacing="0" cellpadding="0">
    <tr><td class="hr"><td></tr>
    <tr><td class="footer">Generated by: <a href="http://gcovr.com">GCOVR (Version {{VERSION}})</a></td></tr>
  </table>
  <br>

</body>

</html>

