# -*- coding:utf-8 -*-

# This file is part of gcovr <http://gcovr.com/>.
#
# Copyright 2013-2018 the gcovr authors
# Copyright 2013 Sandia Corporation
# This software is distributed under the BSD license.

# This file contains ONLY the version number.
# It is executed by setup.py to load the current version.
# This module may not depend on other parts of gcovr.

__version__ = "4.2"
